searchData={"items":[{"type":"protocol","title":"ExCop.Policy.Protocol","doc":"","ref":"ExCop.Policy.Protocol.html"},{"type":"function","title":"ExCop.Policy.Protocol.before/6","doc":"","ref":"ExCop.Policy.Protocol.html#before/6"},{"type":"function","title":"ExCop.Policy.Protocol.can?/6","doc":"","ref":"ExCop.Policy.Protocol.html#can?/6"},{"type":"type","title":"ExCop.Policy.Protocol.args/0","doc":"","ref":"ExCop.Policy.Protocol.html#t:args/0"},{"type":"type","title":"ExCop.Policy.Protocol.context/0","doc":"","ref":"ExCop.Policy.Protocol.html#t:context/0"},{"type":"type","title":"ExCop.Policy.Protocol.error_response/0","doc":"","ref":"ExCop.Policy.Protocol.html#t:error_response/0"},{"type":"type","title":"ExCop.Policy.Protocol.everything/0","doc":"","ref":"ExCop.Policy.Protocol.html#t:everything/0"},{"type":"type","title":"ExCop.Policy.Protocol.field/0","doc":"","ref":"ExCop.Policy.Protocol.html#t:field/0"},{"type":"type","title":"ExCop.Policy.Protocol.parent/0","doc":"","ref":"ExCop.Policy.Protocol.html#t:parent/0"},{"type":"type","title":"ExCop.Policy.Protocol.persona/0","doc":"","ref":"ExCop.Policy.Protocol.html#t:persona/0"},{"type":"type","title":"ExCop.Policy.Protocol.response/0","doc":"","ref":"ExCop.Policy.Protocol.html#t:response/0"},{"type":"type","title":"ExCop.Policy.Protocol.t/0","doc":"All the types that implement this protocol.","ref":"ExCop.Policy.Protocol.html#t:t/0"},{"type":"extras","title":"ExCop","doc":"# ExCop\n\nExCop is a flexible policy DSL allowing you to write easy-to-navigate policies. ExCop doesn't make any assumption\nabout what kind of layer is on top or under it - and even though it was designed to work particularily easily\nwith Absinthe, it should also work for many other environments.","ref":"readme.html"},{"type":"extras","title":"Installation - ExCop","doc":"If [available in Hex](https://hex.pm/docs/publish), the package can be installed\nby adding `ex_cop` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:ex_cop, \"~> 0.1.0\"}\n  ]\nend\n```\n\nDocumentation can be generated with [ExDoc](https://github.com/elixir-lang/ex_doc)\nand published on [HexDocs](https://hexdocs.pm). Once published, the docs can\nbe found at [https://hexdocs.pm/ex_cop](https://hexdocs.pm/ex_cop).","ref":"readme.html#installation"},{"type":"extras","title":"Configuration - ExCop","doc":"You probably want to declare a `Police` module that looks like this:\n\n```elixir\ndefmodule MyApp.Police do\n  @moduledoc false\n  alias ExCop.Policy.Protocol\n  import ExCop.Police, only: [allow: 0]\n\n  @type user :: Protocol.user()\n  @type error_response :: Protocol.error_response()\n  @type response :: Protocol.response()\n  @type parent :: Protocol.parent()\n  @type field :: Protocol.field()\n  @type context :: Protocol.context()\n  @type args :: Protocol.args()\n\n  @allowed_parents [\n    :__schema,\n    :__directive,\n    :__type,\n    :__inputvalue,\n    :__field,\n    :__enumvalue,\n    :page_info,\n    :edges\n  ]\n\n  @allowed_fields [\n    :__typename,\n    :__schema,\n    :id,\n    :node,\n    :nodes,\n    :edges,\n    :cursor,\n    :page_info\n  ]\n\n  @spec check(any, user, parent, field, context, args) :: response()\n  def check(_source, _user, parent, field, _ctx, _args)\n      when parent in @allowed_parents or field in @allowed_fields,\n      do: allow()\n\n  def check(source, user, parent, field, ctx, args) do\n    parent\n    |> to_string\n    |> String.ends_with?(\"_payload\")\n    |> case do\n      true -> allow()\n      _ -> source |> ExCop.Police.check(user, parent, field, ctx, args)\n    end\n  end\n\n  defmodule Helpers do\n    @moduledoc false\n\n    defmacro object_allowance(title, do: block) do\n      quote do\n        allowance unquote(title) do\n          unquote(block)\n          guard var!(parent) not in [:query, :mutation, :subscription]\n        end\n      end\n    end\n\n    # A macro that requires a user as the current persona.\n    defmacro requires_logged_in_user() do\n      quote do\n        persona %User{}\n      end\n    end\n\n    # A macro that requires an admin user as the current persona.\n    defmacro requires_admin_user() do\n      quote do\n        persona %User{is_admin: true}\n      end\n    end\n\n    # A macro that requires that the current persona is `nil` - some guest systems do that.\n    defmacro requires_guest_user() do\n      quote do\n        persona nil\n      end\n    end\n  end\nend\n```\n\nThen you could have a `Policy` module defined such as:\n\n```elixir\ndefmodule MyApp.Policy do\n  @moduledoc false\n\n  defmacro __using__(opts) do\n    quote do\n      use ExCop.Policy, unquote(opts)\n      alias MyApp.Schema.Accounts.{User, Session}\n      import MyApp.Police.Helpers\n    end\n  end\nend\n```\n\nThis way, all the macro you've defined in MyApp.Police.Helpers would become available to your policies when they `use MyApp.Policy, target: Something`.","ref":"readme.html#configuration"},{"type":"extras","title":"Policy System - ExCop","doc":"To write a policy, simply create a module looking like the following:\n\n```elixir\ndefmodule MyApp.PostPolicy do\n  alias MyApp.{Post, User}\n  use MyApp.Policy, target: Post\n\n  allowance \"all users can see a post title and body if it's valid\" do\n    # We require that the `%Post{}` subject has a field `valid` set to `true`.\n    subject %{valid: true}\n    field_in [:title, :body]\n  end\n\n  allowance \"users can see the author of a post\" do\n    # Equivalent to `persona %User{}`...\n    requires_logged_in_user()\n    field :author\n  end\n\n  allowance \"users can see everything on posts they authored\" do\n    persona %User{id: user_id}\n    subject %{author_id: user_id}\n  end\n\n  allowance \"posts with less than three comments can be seen by users\" do\n    # Here, we guard against the shape of a particular subject, and later use that binding.\n    requires_logged_in_user()\n    subject %{comment_count: count}\n    guard count < 3\n  end\n\n  allowance \"allows CIA users to see everything on posts for area 51\" do\n    subject %{cia_post: true, mission: mission}\n    check do\n      String.downcase(persona.agency) == \"cia\" && mission =~ \"area 51\"\n    end\n  end\n\n  allowance \"administrators can see everything in a post\" do\n    requires_admin_user()\n  end\nend\n```\n\nIf you're using Absinthe and want to control what is happening at the root of your schema, you'll have to\nimplement a policy such as this one:\n\n```elixir\ndefmodule MyApp.RootPolicy do\n  @moduledoc false\n  use MyApp.Policy, target: Map\n\n  # Shortcut to using `parent :query`.\n  query_allowance \"users can access certain queries\" do\n    persona %User{}\n    field_in [:me, :users, :onboards, :documents]\n  end\n\n  # Shortcut to using `parent :mutation`.\n  mutation_allowance \"guests can create new users and authenticate\" do\n    persona nil\n    field_in [:create_user, :authenticate]\n  end\nend\n```","ref":"readme.html#policy-system"},{"type":"extras","title":"Policy Delegation to Another Policy - ExCop","doc":"Another trick you can leverage while using ExCop is the policy delegation feature. Consider something like\nthe following:\n\n```elixir\ndefmodule MyApp.RootPolicy do\n  @moduledoc false\n  use MyApp.Policy, target: Map\n\n  # Shortcut to using `parent :query`.\n  mutation_allowance \"users can access certain queries\" do\n    field :add_comment\n    delegated()\n  end\nend\n```\n\nThis particular policy would destructure the `context` into `%{fetched: %{subject: subject}}` and call\n`ExCop.Policy.Protocol.can?/6`, effectivelly replacing the subject by the one found in context. Therefore,\nif the subject is of a different type, the protocol would in turn try to find a matching policy for the new\nsubject.\n\n> Note that for this mechanism to work, you will need to have the context fetched before you try and apply\n> policies to your subject. See the \"Loading Subjects\" paragraph for more information.","ref":"readme.html#policy-delegation-to-another-policy"},{"type":"extras","title":"Advices - ExCop","doc":"Policies are compiled into Elixir. A module will be created conforming to the `ExCop.Policy` and\ndeclaring a series of `can?/6` functions, one per `policy` you called.\n\nAt runtime, policies are evaluated from top to bottom - so it might be a good idea to keep the most used ones\non top and the most expensive ones to run at the bottom, like you would usually do with pattern-matching.\n\nIn case of a default:\n- If no policy module exists for the target, it will return `{:error, :missing_policy}`.\n- When no policy is found matching the arguments, `{:error, :unauthorized}` will be returned.\n\nNote that most of the builder is defined using macros - so you will get compile-time errors and warnings\nif you define a guard using a variable that is unknown, or if you declare a binding that is not used.\n\nFor more informations about how policies can be written, please check `test/ex_cop/police_test.exs` and\nits fixtures.","ref":"readme.html#advices"},{"type":"extras","title":"The `check/0` Function - ExCop","doc":"If you decide to use the `check` function, be mindful of the following:\n\n- You can use the following bindings `subject`, `user`, `parent`, `field`, `context` and `args`.\n- Your check block will allow access if it returns anything else than `false`-ey.\n- The check function is performed after pattern matching - meaning that once the check block is entered, no further policy will be evaluated, even if the check block returns `false`-ey.","ref":"readme.html#the-check-0-function"},{"type":"extras","title":"Manually Verifying Policies - ExCop","doc":"To check for a policy, you can do something like this:\n\n```elixir\nsource |> ExCop.Police.check(user, parent, field, context, args)\n```","ref":"readme.html#manually-verifying-policies"},{"type":"extras","title":"Loading Subjects - ExCop","doc":"For certain policies, you want to make sure that the subject is loaded before the policies are ran. In the\ncase of Absinthe, it means that you might want to have your `subject` loaded before your authorization layer\nkicks-in.","ref":"readme.html#loading-subjects"}],"content_type":"text/markdown"}